/**************************************************************************
 * File		: include/linux/ktau/ktau_proc_int.h
 * Version 	: $Id: ktau_proc_int.h,v 1.5 2006/11/12 00:26:28 anataraj Exp $
 *************************************************************************/
#ifndef _KTAU_PROC_INT_H
#define _KTAU_PROC_INT_H


/* /proc/ktau internal kernel stuff */

#include <linux/types.h>

#define TAU_PROC_ROOT "ktau"
#define TAU_PROC_MAX_NAME 256

/* read, write, ioctl routines registerted for the /proc/ktau/... files */
extern int ktau_read_proc(char *page, char **start, off_t off, 
		int count, int *eof, void *data);

extern int ktau_write_proc(struct file* file, const char* buffer, 
		unsigned long count, void* data);

static int ktau_ioctl_proc(struct inode *inode, struct file *file,
		unsigned int cmd, unsigned long arg);
#if 0
extern int ktau_read_all(char *page, char **start, off_t off, 
		int count, int *eof, void *data);

extern int ktau_write_all(struct file* file, const char* buffer, 
		unsigned long count, void* data);

static int ktau_ioctl_all(struct inode *inode, struct file *file,
		unsigned int cmd, unsigned long arg);
#endif /* 0 */

extern struct file_operations tau_fops;



/* Helpers */
static int do_ktau_merge(void* data, unsigned int cmd, ktau_merge_t *ktau_merge, int self, unsigned int hdr_size);

static int do_ktau_size(void* data, pid_t* pidlist, int nopids, unsigned int cmd, ktau_size_t *ktau_sizei, int self, unsigned int hdr_size);

static int do_ktau_read(void* data, pid_t* pidlist, int nopids, unsigned int cmd, ktau_read_t *ktau_read, int self, unsigned int hdr_size);

/*static int do_ktau_write(void* data, unsigned int cmd, ktau_write_t *ktau_wirte, int self, unsigned int hdr_size);*/

/* Used to collapse the dumpHash output into a single large array of bytes to transfer to user-space */
static int alloc_and_collapse(char** buf, ktau_output* data, unsigned int noprofs, unsigned int hdr_size);

/* a seperate _do_read, as its currently used by both do_size & do_read */
static int _do_ktau_read(void* data, pid_t* pidlist, int nopids, unsigned int cmd, ktau_read_t *ktau_read, int self, unsigned int hdr_size);

/* we need to know how to free the complex-data-struct returned from dumpHash.  *
 * This function has the logic to do so. Ideally it should be held by ktau, not *
 * ktau_proc */
static void free_profile_data(ktau_output* profile_data, int noprofs);

#ifdef CONFIG_KTAU_TRACE
static int do_ktau_trace_size(void* data, pid_t* pidlist, int nopids, unsigned int cmd, ktau_size_t *ktau_sizei, int self, unsigned int hdr_size);

static int do_ktau_trace_read(void* data, pid_t* pidlist, int nopids, unsigned int cmd, ktau_read_t *ktau_read, int self, unsigned int hdr_size);

/* Used to collapse the dumpHash output into a single large array of bytes to transfer to user-space */
static int alloc_and_collapse_trace(char** buf, ktau_output* data, unsigned int noprofs, unsigned int hdr_size);

/* a seperate _do_read, as its currently used by both do_size & do_read */
static int _do_ktau_trace_read(void* data, pid_t* pidlist, int nopids, unsigned int cmd, ktau_read_t *ktau_read, int self, unsigned int hdr_size);

/* we need to know how to free the complex-data-struct returned from dumpHash.  *
 * This function has the logic to do so. Ideally it should be held by ktau, not *
 * ktau_proc */
static void free_trace_data(ktau_output* profile_data, int noprofs);
#endif /*CONFIG_KTAU_TRACE*/

/* ... */
static int get_usrpidlist(pid_t** ppidlist, int nopids, char* user_buf);

/* ... */
static void free_usrpidlist(pid_t** ppidlist);

#endif /* _KTAU_PROC_INT_H */
