/*************************************************************
 * File 	: kernel/ktau/ktau_hash.c
 * Version 	: $Id: ktau_hash.c,v 1.11 2006/11/12 00:26:29 anataraj Exp $
 *
 ************************************************************/

#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/init.h>
#include <linux/linkage.h>      /* system call stuff */
#include <linux/sched.h>        /* task_struct def. */
#include <linux/slab.h>		/* kmalloc() */
#include <linux/interrupt.h>
#include <linux/spinlock.h>     /* spinlock */
#include <linux/rwsem.h>        /* R/W semaphore */
//#include <linux/hardirq.h>	/* This might be needed for 2.6 */

#include <asm/current.h>	/* get_current */
#include <asm/uaccess.h>

#define TAU_NAME "ktau_hash"
#define PFX TAU_NAME

//#define TAU_DEBUG		/* to turn on debug messages */
#include <linux/ktau/ktau_print.h>
#include <linux/ktau/ktau_hash.h>
#include <linux/ktau/ktau_api.h>
#include <linux/ktau/ktau_proc.h>
#include <linux/ktau/ktau_mmap.h>
#include <linux/ktau/ktau_bootopt.h>

#include "ktau_linux-2.4.c"

extern rwlock_t tasklist_lock;

/*----------------------- KTAU GLOBAL SEMAPHORE -----------------------*/
struct rw_semaphore ktau_global_sem;
void ktau_global_semaphore_init(void){
	init_rwsem(&ktau_global_sem);
}

/*--------------------------- IN API ---------------------------------*/
/* only to be called on tasks which are about to be removed - i.e. the "fromtsk" should be dead */
void transfer_task_profile(struct task_struct *fromtask, struct task_struct *totask) {
        //from is usually a child
        //to is usually the parent
        unsigned long kflags = 0;
        int nocopied = 0, true_frmcount = 0, i = 0;
        h_ent *from_h_ptr, *to_h_ptr;
        ktau_prof *frombuf, *tobuf;

        if((!fromtask) || (!totask)) {
                err("transfer_task_profile: null tasks: from:%p  to:%p\n", fromtask, totask);
                return;
        }

        frombuf = fromtask->ktau;
        tobuf = totask->ktau;
        true_frmcount = frombuf->ent_count;

        //is it sufficient to lock only "to" ,as "from" tsk is dead?
        ktau_spin_lock(&totask->ktau_lock,kflags);

        /* Searching through hash table */
        for(i=0;i<KTAU_HASH_SIZE;i++){
                from_h_ptr = (h_ent*)(frombuf->ktau_hash[i]);
                if(from_h_ptr){

                        to_h_ptr = (h_ent*)(tobuf->ktau_hash[i]);
			if(!to_h_ptr) {
				//AN_20060419_OPTIMZ
				//current->ktau->ktau_hash[index] = kmalloc(sizeof(h_ent),GFP_ATOMIC);
				if(!tobuf->hash_mem) {
					goto ktau_unlock;
				}
				tobuf->ktau_hash[i] = (tobuf->hash_mem+i);

                                *(tobuf->ktau_hash[i]) = *from_h_ptr;

                                //init other state inside h_ptr
                                tobuf->ktau_hash[i]->nestingLevel = 0;
                                tobuf->ktau_hash[i]->data.timer.start = 0;

                                tobuf->ent_count++;
			} else {               
/*
                        if(!to_h_ptr) {
                                //then just hand over the memory
                                tobuf->ktau_hash[i] = from_h_ptr;
                                frombuf->ktau_hash[i] = NULL;
                                frombuf->ent_count--;
                                tobuf->ent_count++;

                                //init other state inside h_ptr
                                tobuf->ktau_hash[i]->nestingLevel = 0;
                                tobuf->ktau_hash[i]->data.timer.start = 0;
                        } else {
*/
                                //otherwise sum it
                                to_h_ptr->data.timer.count += from_h_ptr->data.timer.count;
                                to_h_ptr->data.timer.incl += from_h_ptr->data.timer.incl;
                                to_h_ptr->data.timer.excl += from_h_ptr->data.timer.excl;
                        }

                        nocopied++;

                        /* Check if all data are coppied over. */
                        if(nocopied >= true_frmcount)
                                break;
                }
        }
ktau_unlock:
        ktau_spin_unlock(&totask->ktau_lock,kflags);
}

/*
 * This function initialize ktau_prof structure in task_struct.
 * 	1. Allocate the hash-table which contains pointer 
 * 	   to profile information (h_ent).
 * 	2. Initialize spinlock.
 * 	3. Create the corresponding entry in /proc/<pid>/ktau.
 */
void create_task_profile(struct task_struct *tsk){

	//AN_20060419_OPTIMZ
	/*
	if(!(tsk->ktau = kmalloc(sizeof(ktau_prof),GFP_KERNEL))){
                err("Can't allocate memory for ktau_prof.");
	}else{
		memset(tsk->ktau->ktau_hash,0,KTAU_HASH_SIZE * sizeof(h_ent*));

		tsk->ktau->hash_mem = 0;

		if(!(tsk->ktau->hash_mem = kmalloc(sizeof(h_ent) * KTAU_HASH_SIZE, GFP_KERNEL))) {
			err("Can't allocate memory for ktau_prof->hash_mem: size:%d\n", sizeof(h_ent) * KTAU_HASH_SIZE);
		}
	}
	*/
	//AN_20060419_OPTIMZ_2
	//calculate total space required for the profile and allocate it
	char* tot_space = NULL;
	int tot_num = 0;

#ifndef CONFIG_KTAU_TRACE /* not defined TRACE -then do profile */
	tot_num = (/*for tsk->ktau*/ sizeof(ktau_prof)) +  (/*for h_ents*/ sizeof(h_ent) * KTAU_HASH_SIZE); 
	if(!(tot_space = kmalloc(tot_num, GFP_KERNEL))) {
                err("Can't allocate total memory for ktau_prof.");
		tsk->ktau = NULL;
		return;
	}
	//assign to the different entities from the tot_space
	tsk->ktau = (ktau_prof*) tot_space;
	tot_space += sizeof(ktau_prof);
	memset(tsk->ktau->ktau_hash,0,KTAU_HASH_SIZE * sizeof(h_ent*));
	tsk->ktau->hash_mem = (h_ent*) tot_space;
	//done assignment of memory -----
	
	/* Initialize entry counter */
	tsk->ktau->ent_count 	= 0;
	memset(&tsk->ktau->accum,0,
			sizeof(unsigned long long)*KTAU_NESTED_LEVEL+1);
	//FIX MERGE
	memset(&(tsk->ktau->merge), 0, sizeof(ktau_user));

	//AN_20051102_REENT
	memset(&tsk->ktau->reent,0, sizeof(unsigned long long)*(KTAU_NESTED_LEVEL+1));
	tsk->ktau->stackDepth = 0;
	tsk->ktau->last_sysc = -(0xAEEE);

#else /*this means Trace is defined - therefore dont do profile*/

	tot_num = (/*for tsk->ktau*/ sizeof(ktau_prof));
	if(!(tot_space = kmalloc(tot_num, GFP_KERNEL))) {
                err("Can't allocate total memory for ktau_prof.");
		tsk->ktau = NULL;
		return;
	}
	//assign to the different entities from the tot_space
	tsk->ktau = (ktau_prof*) tot_space;
	//memset(tsk->ktau->ktau_hash,0,KTAU_HASH_SIZE * sizeof(h_ent*));
	tsk->ktau->hash_mem = NULL;
	//done assignment of memory -----

	//must do this for both trace and profile
	tsk->ktau->last_sysc = -(0xAEEE);

#ifdef CONFIG_KTAU_BOOTOPT
	if(ktau_verify_bootopt(KTAU_TRACE_MSK))
#endif /* CONFIG_KTAU_BOOTOPT*/
	{
		tsk->ktau->trbuf	= NULL;
		tsk->ktau->cur_trace	= NULL;
		tsk->ktau->head_trace	= NULL;
		trace_buf_init(tsk);	
	}
#endif /*CONFIG_KTAU_TRACE*/

	/* Initialize hash table lock */
	spin_lock_init(&tsk->ktau_lock);

	/* Point internal hash-lock to ktau_lock in task_struct */
	/* - so lots of code doesn't break. Should change rest  */
	/* of code that breaks... */
	tsk->ktau->hash_lock = &(tsk->ktau_lock);

}

/*
 * This function handles clean up for ktau_prof structure
 * in task_struct.
 * 	1. Freeing all the allocated profiles.
 */
void remove_task_profile(struct task_struct *tsk){
	unsigned long kflags = 0;
#ifdef CONFIG_KTAU_TRACE
	void *temp_trace = NULL;
#endif /*CONFIG_KTAU_TRACE*/

	/* DONT DO this parent thing - typically unnecessary - make it an option
	if((tsk->parent)  && (tsk->parent->ktau)) {
		transfer_task_profile(tsk, tsk->parent);
	}
	*/

	ktau_spin_lock(&tsk->ktau_lock,kflags);

	//AN_20060419_OPTIMZ
	/* Free the memory allocated as a whole in hash_mem */
	/*
	if(likely(tsk->ktau->hash_mem)) {
		//AN_20060419_OPTIMZ_2
		//kfree(tsk->ktau->hash_mem);
		tsk->ktau->hash_mem = NULL;
	}
	*/
	tsk->ktau->hash_mem = NULL;

	/* Free the memory allocated for each entry */
	/*
	for(i=0;i<KTAU_HASH_SIZE;i++){
		if((tsk->ktau->ktau_hash[i])){
			//AN_20060419_OPTIMZ
			//kfree((tsk->ktau->ktau_hash[i]));
			tsk->ktau->ktau_hash[i] = NULL;	
		}
        }
	*/

#ifndef CONFIG_KTAU_TRACE
	if(unlikely(tsk->ktau->merge.uptr != 0)) {
	       //TODO: Must get rid of the mmaped 'merge' pages	
	       //the process died before it cleaned up properly.
	       err("remove_task_profile: WARN: merge.uptr != NULL. pid:%d\n", tsk->pid);	
	}
#endif

#ifdef CONFIG_KTAU_TRACE
#ifdef CONFIG_KTAU_BOOTOPT
	if(ktau_verify_bootopt(KTAU_TRACE_MSK))
#endif /* CONFIG_KTAU_BOOTOPT*/
	{
		/* Free ktau_prof */
		temp_trace = trace_buf_destroy(tsk);
	}
#endif /*CONFIG_KTAU_TRACE*/

	kfree(tsk->ktau);
	tsk->ktau = NULL;

	ktau_spin_unlock(&tsk->ktau_lock,kflags);

#ifdef CONFIG_KTAU_TRACE
	trace_buf_free(temp_trace);
#endif /*CONFIG_KTAU_TRACE*/

}

/*--------------------------- OUT API ---------------------------------*/
/*
 * This function handles dumping the profile hash-table
 */
int dump_hash_self(struct task_struct* tsk, ktau_prof *buf, 
		ktau_output *out, 
		unsigned long long tol)
{
        int i;
	unsigned long kflags = 0;
	h_ent* h_ptr; 
	o_ent* o_ptr; 

	if(out == NULL)
		return -1;

	out->size = 0;
	out->ent_lst = kmalloc(sizeof(o_ent) * buf->ent_count , GFP_KERNEL);

	ktau_spin_lock(&tsk->ktau_lock,kflags);

	/* Searching through hash table */
	for(i=0;i<KTAU_HASH_SIZE;i++){
		h_ptr = (h_ent*)(buf->ktau_hash[i]);
		if(h_ptr){
						
			/* Copy data over */
			o_ptr = (out->ent_lst + out->size);
			o_ptr->index = i;
			memcpy(&(o_ptr->entry),h_ptr, sizeof(h_ent));
			
			/*
			printk(KERN_INFO "dump_hash_self: Entry %4d: addr %x, count %4u, start %4llu, incl %4llu, excl %4llu\n",
				i,
				o_ptr->entry.addr, 
				o_ptr->entry.data.timer.count,
				o_ptr->entry.data.timer.start,
				o_ptr->entry.data.timer.incl,
				o_ptr->entry.data.timer.excl);
			*/		
			out->size++;

			/* Check if all data are coppied over. */
			if(out->size >= buf->ent_count) 
				break;
		}
        }

	ktau_spin_unlock(&tsk->ktau_lock,kflags);
	return 0;
}

/* 
 * This function returns number of entries for 
 * "self".
 */ 
int get_profile_size_self(ktau_prof *buf)
{
	int count	= 0;
	int i		= 0;

	for(i=0;i<KTAU_HASH_SIZE;i++){
		if(buf->ktau_hash[i])
			count++;
	}
	
	return(count);
}

/* 
 * This function returns number of entries for 
 * "many".
 */ 
int get_profile_size_many(pid_t *pid_lst, int* num_pid)
{
	struct task_struct *cur_tsk;
	int count	= 0;
	int i		= 0;
	int cur_count   = 0;
	
	/* Case for all */
	if(*num_pid == 0){
		read_lock(&tasklist_lock);
		for_each_process(cur_tsk){
			*num_pid = (*num_pid) + 1;
			for(i=0;i<KTAU_HASH_SIZE;i++){
				if(cur_tsk->ktau->ktau_hash[i])
					count++;
			}
		}
		read_unlock(&tasklist_lock);	
	}
	/* Case for many pids */	
	else{
		read_lock(&tasklist_lock);
#if 0
		for_each_process(cur_tsk){
			if(cur_tsk->pid == pid_lst[cur_count]){
				/* The specified PID is found */
				for(i=0;i<KTAU_HASH_SIZE;i++){
					if(cur_tsk->ktau->ktau_hash[i])
						count++;
				}
				cur_count++;
			}else if(cur_tsk->pid > pid_lst[cur_count]){
				/* The specified PID is not found */
				cur_count++;
			}
		}
#endif
		for_each_process(cur_tsk){
			do{
				if(cur_tsk->pid == pid_lst[cur_count]){
					/* The specified PID is found */
					
					for(i=0;i<KTAU_HASH_SIZE;i++){
						if(cur_tsk->ktau->ktau_hash[i])
							count++;
					}

					/* Move to the next pid in the list */
					cur_count++;
				
				}else if(cur_tsk->pid > pid_lst[cur_count]){
					/* The specified PID is not found */
					/* Move to the next pid in the list */
					err("get_profile_size_many : WARNING! PID %d not found.\n",
							pid_lst[cur_count]);	
					cur_count++;
				}
				/* Must check for array out-of-bound 
				 * before continue */
				if(cur_count >=  *num_pid) 
					goto done;

			}while(cur_tsk->pid >= pid_lst[cur_count]);
		}
done:
		read_unlock(&tasklist_lock);	
	}
	return(count);
}
/*
 * This function handles purging the profile hash-table
 */
int purge_hash_self(ktau_prof *buf, 
		unsigned long long tol)
{
#if 0
        int i;
	unsigned long flags = 0;
	h_ent* h_ptr;

	ktau_spin_lock(&buf->hash_lock,flags);
		for(i=0;i<KTAU_HASH_SIZE;i++){
			h_ptr = (h_ent*)(buf->ktau_hash[i]);
			if(h_ptr){
				h_ptr->data.timer.count = 0;	
				h_ptr->data.timer.start = 0;	
				h_ptr->data.timer.incl = 0;
				h_ptr->data.timer.excl = 0;
			}
		}
	ktau_spin_unlock(&buf->hash_lock,flags);
#endif
	return 0;
}
/*-------------------------------------------------------------------*/
#if 0 
/* Replaced by dump_hash_many and purge_hash_many */
/*
 * Wrapper of dump_hash_self w/ locking 
 */
int dump_hash_other(ktau_prof *buf, 
		ktau_output *out, 
		unsigned long long tol)
{
	int ret;
	/*
	 * First: Acquire the global semaphore and tasklist_lock
	 */
	down_write(&ktau_global_sem);
	read_lock(&tasklist_lock);

	ret = dump_hash_self(buf,out,tol);

	/*Last : Release the global semaphore and tasklist_lock*/
	read_unlock(&tasklist_lock);
	up_write(&ktau_global_sem);
	return ret;
}

/*
 * Wrapper of purge_hash_self w/ locking
 */
int purge_hash_other(ktau_prof *buf, 
		unsigned long long tol)
{
	int ret;
	down_write(&ktau_global_sem);
	read_lock(&tasklist_lock);

	ret = purge_hash_self(buf,tol);

	/*Last : Release the global semaphore and tasklist_lock*/
	read_unlock(&tasklist_lock);
	up_write(&ktau_global_sem);
	return ret;
}
#endif 
/*-------------------------------------------------------------------*/
/*
 * This is wrapped by dump_hash_many in case we are dumping all.
 */

static int _dump_hash_all(ktau_output **arr,
		unsigned long long tol)
{
	struct task_struct *cur_tsk;
	int count = 0;
	int cur_count= 0;

	down_write(&ktau_global_sem);

	read_lock(&tasklist_lock);

	for_each_process(cur_tsk) count++;

	*arr = kmalloc(sizeof(ktau_output) * count, GFP_ATOMIC);	

	for_each_process(cur_tsk){
		(*arr+cur_count)->pid = cur_tsk->pid;
		if(dump_hash_self(cur_tsk, cur_tsk->ktau, (*arr+cur_count), tol))
			break;
		cur_count++;
	}
	read_unlock(&tasklist_lock);
	up_write(&ktau_global_sem);

	return cur_count;
}

/*
 * This is wrapped by purge_hash_all in case we are purging all. 
 */
#if 0
static int _purge_hash_all(unsigned long long tol)
{
	int cur_count = 0;
	struct task_struct *cur_tsk;

	down_write(&ktau_global_sem);
	read_lock(&tasklist_lock);
	for_each_process(cur_tsk){
		if(purge_hash_self(cur_tsk->ktau, tol))
			break;
		cur_count++;
	}

	read_unlock(&tasklist_lock);
	up_write(&ktau_global_sem);

	return cur_count;
}
#endif
/*-------------------------------------------------------------------*/
/*
 * dump_hash_many: Perform dump on other and all.
 */
int dump_hash_many(pid_t *pid_lst,
		unsigned int size, 
		ktau_output **arr, 
		unsigned long long tol)
{
	int ret_count = 0;
	int cur_count = 0;
	struct task_struct *cur_tsk;

	/* Check if passing ALL */
	if(!size){
		if(arr) return _dump_hash_all(arr,tol);
		else    return(-1);		/* Error arr is NULL */
	}else
		*arr = kmalloc(sizeof(ktau_output)*size,GFP_KERNEL);
	
	down_write(&ktau_global_sem);
	read_lock(&tasklist_lock);

	/* 
	 * Searching through the task list for the specified process.
	 * This should be optimized!!!!! Required the pid_list to be 
	 * sorted.
	 */

	for_each_process(cur_tsk){
		do{
			if(cur_tsk->pid == pid_lst[cur_count]){
				/* The specified PID is found */
				
				/* Read the pid from the process task_struct */	
				(*arr+ret_count)->pid = cur_tsk->pid;
			
				/* Do the same thing as dump_hash_self */
				if(dump_hash_self(cur_tsk, cur_tsk->ktau, (*arr+ret_count), tol))
					break;

				/* Move to the next pid in the list */
				cur_count++;
				ret_count++;
			
			}else if(cur_tsk->pid > pid_lst[cur_count]){
				/* The specified PID is not found */
				/* Move to the next pid in the list */
				err("dump_hash_many : WARNING! PID %d not found.\n",
						pid_lst[cur_count]);	
				cur_count++;
			}
			/* Must check for array out-of-bound 
			 * before continue */
			if(cur_count >=  size || ret_count == size) 
				goto done;

		}while(cur_tsk->pid >= pid_lst[cur_count]);

	}
done:	
	read_unlock(&tasklist_lock);
	up_write(&ktau_global_sem);
	return ret_count;
}
/*
 * purge_hash_many: Perform purge on other and all.
 */
int purge_hash_many(pid_t *pid_lst,
		unsigned int size, 
		unsigned long long tol)
{
	int ret_count = 0;
#if 0
	int cur_count = 0;
	struct task_struct *cur_tsk;

	/* Check if passing ALL */
	if(!size) 
		return _purge_hash_all(tol);


	down_write(&ktau_global_sem);
	read_lock(&tasklist_lock);

	/* 
	 * Searching through the task list for the specified process.
	 * This should be optimized!!!!! Required the pid_list to be 
	 * sorted.
	 */
	for_each_process(cur_tsk){
		if(cur_tsk->pid == pid_lst[cur_count]){
			/* The specified PID is found */
			if(purge_hash_self(cur_tsk->ktau,tol))
				break;
			cur_count++;
			ret_count++;
		}else if(cur_tsk->pid > pid_lst[cur_count]){
			/* The specified PID is not found */
			cur_count++;
		}

		/* Check if we have worked  upto the size specified*/
		if(ret_count == size)
			break;
	}

	read_unlock(&tasklist_lock);
	up_write(&ktau_global_sem);

#endif
	return ret_count;
}
/*--------------------------- CONTROL API ---------------------------------*/
/*  Merging changed & moved to ktau_merge.c / ktau_merge.h */
